@@ -1,6 +1,6 @@
-// Copyright (c) 2020  Brendan Molloy <brendan@bbqsrc.net>,
-//                     Ilya Solovyiov <ilya.solovyiov@gmail.com>,
-//                     Kai Ren <tyranron@gmail.com>
+// Copyright (c) 2020-2022  Brendan Molloy <brendan@bbqsrc.net>,
+//                          Ilya Solovyiov <ilya.solovyiov@gmail.com>,
+//                          Kai Ren <tyranron@gmail.com>
 //
 // Licensed under the Apache License, Version 2.0 <LICENSE-APACHE or
 // http://www.apache.org/licenses/LICENSE-2.0> or the MIT license
@@ -32,6 +32,7 @@
     clippy::decimal_literal_representation,
     clippy::else_if_without_else,
     clippy::empty_line_after_outer_attr,
+    clippy::equatable_if_let,
     clippy::exit,
     clippy::expect_used,
     clippy::fallible_impl_from,
@@ -57,6 +58,7 @@
     clippy::rc_buffer,
     clippy::rc_mutex,
     clippy::rest_pat_in_fully_bound_structs,
+    clippy::same_name_method,
     clippy::shadow_unrelated,
     clippy::str_to_string,
     clippy::string_add,
@@ -94,7 +96,8 @@
 )]
 
 mod attribute;
-mod derive;
+mod parameter;
+mod world_init;
 
 use proc_macro::TokenStream;
 
@@ -115,7 +118,7 @@ macro_rules! step_attribute {
         /// # use std::{convert::Infallible};
         /// #
         /// # use async_trait::async_trait;
-        /// use cucumber::{given, World, WorldInit};
+        /// use cucumber::{given, when, World, WorldInit};
         ///
         /// #[derive(Debug, WorldInit)]
         /// struct MyWorld;
@@ -125,11 +128,12 @@ macro_rules! step_attribute {
         ///     type Error = Infallible;
         ///
         ///     async fn new() -> Result<Self, Self::Error> {
-        ///         Ok(Self {})
+        ///         Ok(Self)
         ///     }
         /// }
         ///
         /// #[given(regex = r"(\S+) is (\d+)")]
+        /// #[when(expr = "{word} is {int}")]
         /// fn test(w: &mut MyWorld, param: String, num: i32) {
         ///     assert_eq!(param, "foo");
         ///     assert_eq!(num, 0);
@@ -141,7 +145,24 @@ macro_rules! step_attribute {
         /// }
         /// ```
         ///
-        /// # Arguments
+        /// # Attribute arguments
+        ///
+        /// - `#[given(regex = "regex")]`
+        ///
+        ///   Uses [`Regex`] for matching the step. [`Regex`] is checked at
+        ///   compile time to have valid syntax.
+        ///
+        /// - `#[given(expr = "cucumber-expression")]`
+        ///
+        ///   Uses [Cucumber Expression][1] for matching the step. It's checked
+        ///   at compile time to have valid syntax.
+        ///
+        /// - `#[given("literal")]`
+        ///
+        ///   Matches the step with an **exact** literal only. Doesn't allow any
+        ///   values capturing to use as function arguments.
+        ///
+        /// # Function arguments
         ///
         /// - First argument has to be mutable reference to the [`WorldInit`]
         ///   deriver (your [`World`] implementer).
@@ -164,7 +185,7 @@ macro_rules! step_attribute {
         /// #     type Error = Infallible;
         /// #
         /// #     async fn new() -> Result<Self, Self::Error> {
-        /// #         Ok(Self {})
+        /// #         Ok(Self)
         /// #     }
         /// # }
         /// #
@@ -185,13 +206,22 @@ macro_rules! step_attribute {
         /// # }
         /// ```
         ///
+        /// # Return value
+        ///
+        /// A function may also return a [`Result`], which [`Err`] is expected
+        /// to implement [`Display`], so returning it will cause the step to
+        /// fail.
+        ///
+        /// [`Display`]: std::fmt::Display
         /// [`FromStr`]: std::str::FromStr
+        /// [`Regex`]: regex::Regex
         /// [`gherkin::Step`]: https://bit.ly/3j42hcd
         /// [`World`]: https://bit.ly/3j0aWw7
+        /// [1]: cucumber_expressions
         #[proc_macro_attribute]
         pub fn $name(args: TokenStream, input: TokenStream) -> TokenStream {
             attribute::step(std::stringify!($name), args.into(), input.into())
-                .unwrap_or_else(|e| e.to_compile_error())
+                .unwrap_or_else(syn::Error::into_compile_error)
                 .into()
         }
     };
@@ -208,8 +238,8 @@ macro_rules! steps {
         /// for further details.
         #[proc_macro_derive(WorldInit)]
         pub fn derive_init(input: TokenStream) -> TokenStream {
-            derive::world_init(input.into(), &[$(std::stringify!($name)),*])
-                .unwrap_or_else(|e| e.to_compile_error())
+            world_init::derive(input.into(), &[$(std::stringify!($name)),*])
+                .unwrap_or_else(syn::Error::into_compile_error)
                 .into()
         }
 
@@ -218,3 +248,65 @@ macro_rules! steps {
 }
 
 steps!(given, when, then);
+
+/// In addition to [default parameters] of [Cucumber Expressions], you may
+/// implement and use custom ones.
+///
+/// # Example
+///
+/// ```rust
+/// # use std::{convert::Infallible};
+/// #
+/// # use async_trait::async_trait;
+/// use cucumber::{given, when, Parameter, World, WorldInit};
+/// use derive_more::{Deref, FromStr};
+///
+/// #[derive(Debug, WorldInit)]
+/// struct MyWorld;
+///
+/// #[async_trait(?Send)]
+/// impl World for MyWorld {
+///     type Error = Infallible;
+///
+///     async fn new() -> Result<Self, Self::Error> {
+///         Ok(Self)
+///     }
+/// }
+///
+/// #[given(regex = r"^(\S+) is (\d+)$")]
+/// #[when(expr = "{word} is {u64}")]
+/// fn test(w: &mut MyWorld, param: String, num: CustomU64) {
+///     assert_eq!(param, "foo");
+///     assert_eq!(*num, 0);
+/// }
+///
+/// #[derive(Deref, FromStr, Parameter)]
+/// #[param(regex = r"\d+", name = "u64")]
+/// struct CustomU64(u64);
+/// #
+/// # #[tokio::main]
+/// # async fn main() {
+/// #     MyWorld::run("./tests/features/doctests.feature").await;
+/// # }
+/// ```
+///
+/// # Attribute arguments
+///
+/// - `#[param(regex = "regex")]`
+///
+///   [`Regex`] to match this parameter. Shouldn't contain any capturing groups.
+///
+/// - `#[param(name = "name")]` (optional)
+///
+///   Name of this parameter to reference it by. If not specified, then
+///   lower-cased type name will be used by default.
+///
+/// [`Regex`]: regex::Regex
+/// [Cucumber Expressions]: https://cucumber.github.io/cucumber-expressions
+/// [default parameters]: cucumber_expressions::Expression#parameter-types
+#[proc_macro_derive(Parameter, attributes(param))]
+pub fn parameter(input: TokenStream) -> TokenStream {
+    parameter::derive(input.into())
+        .unwrap_or_else(syn::Error::into_compile_error)
+        .into()
+}
